import logging
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    start_tag_error_text,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib import resource_agent as lib_ra
from pcs.lib.env import LibraryEnvironment

from pcs.lib.commands import stonith_agent as lib


@mock.patch(
    "pcs.lib.resource_agent.list_stonith_agents",
    lambda runner: [
        "fence_apc",
        "fence_dummy",
        "fence_xvm",
    ],
)
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class TestListAgents(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()
        self.lib_env = LibraryEnvironment(self.mock_logger, self.mock_reporter)

    def tearDown(self):
        # pylint: disable=protected-access
        lib_ra.StonithAgent._fenced_metadata = None

    def test_list_all(self):
        self.assertEqual(
            lib.list_agents(self.lib_env, False, None),
            [
                {
                    "name": "fence_apc",
                    "shortdesc": "",
                    "longdesc": "",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [],
                },
                {
                    "name": "fence_dummy",
                    "shortdesc": "",
                    "longdesc": "",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [],
                },
                {
                    "name": "fence_xvm",
                    "shortdesc": "",
                    "longdesc": "",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [],
                },
            ],
        )

    def test_search(self):
        self.assertEqual(
            lib.list_agents(self.lib_env, False, "M"),
            [
                {
                    "name": "fence_dummy",
                    "shortdesc": "",
                    "longdesc": "",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [],
                },
                {
                    "name": "fence_xvm",
                    "shortdesc": "",
                    "longdesc": "",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [],
                },
            ],
        )

    @mock.patch.object(lib_ra.Agent, "_get_metadata", autospec=True)
    def test_describe(self, mock_metadata):
        self.maxDiff = None

        def mock_metadata_func(self):
            if self._get_name() == "ocf:test:Stateful":
                raise lib_ra.UnableToGetAgentMetadata(
                    self._get_name(), "test exception"
                )
            return etree.XML(
                """
                <resource-agent>
                    <shortdesc>short {name}</shortdesc>
                    <longdesc>long {name}</longdesc>
                    <parameters>
                    </parameters>
                    <actions>
                    </actions>
                </resource-agent>
            """.format(
                    name=self._get_name()
                )
            )

        mock_metadata.side_effect = mock_metadata_func

        # Stateful is missing as it does not provide valid metadata - see above
        self.assertEqual(
            lib.list_agents(self.lib_env, True, None),
            [
                {
                    "name": "fence_apc",
                    "shortdesc": "short fence_apc",
                    "longdesc": "long fence_apc",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [
                        {
                            "interval": "60s",
                            "name": "monitor",
                            "OCF_CHECK_LEVEL": None,
                            "automatic": None,
                            "depth": None,
                            "on_target": None,
                            "role": None,
                            "start-delay": None,
                            "timeout": None,
                        }
                    ],
                },
                {
                    "name": "fence_dummy",
                    "shortdesc": "short fence_dummy",
                    "longdesc": "long fence_dummy",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [
                        {
                            "interval": "60s",
                            "name": "monitor",
                            "OCF_CHECK_LEVEL": None,
                            "automatic": None,
                            "depth": None,
                            "on_target": None,
                            "role": None,
                            "start-delay": None,
                            "timeout": None,
                        }
                    ],
                },
                {
                    "name": "fence_xvm",
                    "shortdesc": "short fence_xvm",
                    "longdesc": "long fence_xvm",
                    "parameters": [],
                    "actions": [],
                    "default_actions": [
                        {
                            "interval": "60s",
                            "name": "monitor",
                            "OCF_CHECK_LEVEL": None,
                            "automatic": None,
                            "depth": None,
                            "on_target": None,
                            "role": None,
                            "start-delay": None,
                            "timeout": None,
                        }
                    ],
                },
            ],
        )


@mock.patch.object(lib_ra.StonithAgent, "_load_metadata", autospec=True)
@mock.patch.object(lib_ra.FencedMetadata, "_get_parameters", lambda self: [])
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class TestDescribeAgent(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()
        self.lib_env = LibraryEnvironment(self.mock_logger, self.mock_reporter)
        self.metadata = """
            <resource-agent>
                <shortdesc>short desc</shortdesc>
                <longdesc>long desc</longdesc>
                <parameters>
                </parameters>
                <actions>
                </actions>
            </resource-agent>
        """
        self.description = {
            "name": "fence_dummy",
            "shortdesc": "short desc",
            "longdesc": "long desc",
            "parameters": [],
            "actions": [],
            "default_actions": [
                {
                    "name": "monitor",
                    "interval": "60s",
                    "OCF_CHECK_LEVEL": None,
                    "automatic": None,
                    "depth": None,
                    "on_target": None,
                    "role": None,
                    "start-delay": None,
                    "timeout": None,
                }
            ],
        }

    def tearDown(self):
        # pylint: disable=protected-access
        lib_ra.StonithAgent._fenced_metadata = None

    def test_success(self, mock_metadata):
        mock_metadata.return_value = self.metadata

        self.assertEqual(
            lib.describe_agent(self.lib_env, "fence_dummy"), self.description
        )

        self.assertEqual(len(mock_metadata.mock_calls), 1)

    def test_fail(self, mock_metadata):
        mock_metadata.return_value = "invalid xml"

        assert_raise_library_error(
            lambda: lib.describe_agent(self.lib_env, "fence_dummy"),
            (
                severity.ERROR,
                report_codes.UNABLE_TO_GET_AGENT_METADATA,
                {
                    "agent": "fence_dummy",
                    "reason": start_tag_error_text(),
                },
            ),
        )

        self.assertEqual(len(mock_metadata.mock_calls), 1)



from unittest import TestCase

from pcs_test.tier0.lib.commands.tag.tag_common import (
    fixture_resources_for_ids,
    fixture_tags_xml,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs.lib.commands import tag as cmd_tag


class TestTagConfig(TestCase):
    tag_dicts_list = [
        {
            "tag_id": "tag1",
            "idref_list": ["i1", "i2"],
        },
        {
            "tag_id": "tag2",
            "idref_list": ["j1", "j2"],
        },
    ]

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            tags=fixture_tags_xml(
                [
                    ("tag1", ("i1", "i2")),
                    ("tag2", ("j1", "j2")),
                ]
            ),
            resources=fixture_resources_for_ids(),
        )

    def test_success_no_args(self):
        self.assertEqual(
            cmd_tag.config(self.env_assist.get_env(), []),
            self.tag_dicts_list,
        )

    def test_tag_id_does_not_exist(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.config(
                self.env_assist.get_env(),
                [
                    "nonexistent_tag1",
                    "tag2",
                    "nonexistent_tag2",
                ],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.report_not_found(
                    _id, expected_types=["tag"], context_type="tags"
                )
                for _id in ["nonexistent_tag1", "nonexistent_tag2"]
            ]
        )

    def test_not_a_tag_id(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_tag.config(self.env_assist.get_env(), ["id1"])
        )
        self.env_assist.assert_reports(
            [
                fixture.report_unexpected_element(
                    "id1",
                    "primitive",
                    expected_types=["tag"],
                ),
            ]
        )



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

from pcs.common.reports import codes as report_codes
from pcs.lib.commands import cluster
from pcs.lib.corosync.constants import (
    TOTEM_OPTIONS,
    TRANSPORT_KNET_CRYPTO_OPTIONS,
    TRANSPORT_KNET_COMPRESSION_OPTIONS,
    TRANSPORT_KNET_GENERIC_OPTIONS,
)

from .common import fixture_totem


ALLOWED_COMPRESSION_OPTIONS = list(TRANSPORT_KNET_COMPRESSION_OPTIONS)
ALLOWED_CRYPTO_OPTIONS = list(TRANSPORT_KNET_CRYPTO_OPTIONS)
ALLOWED_TOTEM_OPTIONS = list(TOTEM_OPTIONS)
ALLOWED_KNET_TRANSPORT_OPTIONS = list(TRANSPORT_KNET_GENERIC_OPTIONS)

TRANSPORT_OPTIONS = {
    "ip_version": "ipv4",
    "knet_pmtud_interval": "1234",
    "link_mode": "active",
}
COMPRESSION_OPTIONS = {"level": "5", "model": "zlib", "threshold": "1234"}
CRYPTO_OPTIONS = {"cipher": "aes256", "hash": "sha256", "model": "nss"}
TOTEM_OPTIONS = {
    opt: ("yes" if opt == "block_unlisted_ips" else str(num))
    for num, opt in enumerate(ALLOWED_TOTEM_OPTIONS)
}


class CheckLiveMixin:
    # pylint: disable=invalid-name
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def get_lib_command(self):
        raise NotImplementedError

    def assert_live_required(self, forbidden_options):
        self.env_assist.assert_raise_library_error(
            self.get_lib_command(),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=forbidden_options,
                )
            ],
            expected_in_processor=False,
        )

    def test_mock_corosync(self):
        self.config.env.set_corosync_conf_data("")
        self.assert_live_required(["COROSYNC_CONF"])

    def test_mock_cib(self):
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required(["CIB"])

    def test_mock_cib_corosync(self):
        self.config.env.set_corosync_conf_data("")
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required(["CIB", "COROSYNC_CONF"])


class CheckLiveUpdateConfig(CheckLiveMixin, TestCase):
    def get_lib_command(self):
        return lambda: cluster.config_update(
            self.env_assist.get_env(), {}, {}, {}, {}
        )


class CheckLiveUpdateConfigLocal(CheckLiveMixin, TestCase):
    def get_lib_command(self):
        return lambda: cluster.config_update_local(
            self.env_assist.get_env(), b"", {}, {}, {}, {}
        )


class UpdateConfig(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_no_changed_options(self):
        self.config.corosync_conf.load_content(fixture_totem())
        self.config.env.push_corosync_conf(corosync_conf_text=fixture_totem())
        cluster.config_update(self.env_assist.get_env(), {}, {}, {}, {})
        self.env_assist.assert_reports([])

    def test_add_all_options(self):
        self.config.corosync_conf.load_content(fixture_totem())
        self.config.env.push_corosync_conf(
            corosync_conf_text=fixture_totem(
                transport_options=TRANSPORT_OPTIONS,
                compression_options=COMPRESSION_OPTIONS,
                crypto_options=CRYPTO_OPTIONS,
                totem_options=TOTEM_OPTIONS,
            ),
            need_stopped_cluster=True,
        )
        cluster.config_update(
            self.env_assist.get_env(),
            TRANSPORT_OPTIONS,
            COMPRESSION_OPTIONS,
            CRYPTO_OPTIONS,
            TOTEM_OPTIONS,
        )
        self.env_assist.assert_reports([])

    def _test_modify(self, transport_options, need_stopped_cluster):
        self.config.corosync_conf.load_content(
            fixture_totem(
                transport_options=TRANSPORT_OPTIONS,
                compression_options=COMPRESSION_OPTIONS,
                crypto_options=CRYPTO_OPTIONS,
                totem_options=TOTEM_OPTIONS,
            )
        )
        modified_transport_options = dict(TRANSPORT_OPTIONS)
        modified_transport_options.update(transport_options)
        modified_compression_options = {
            "level": "9",
            "model": "lz4",
            "threshold": "100",
        }
        modified_crypto_options = {
            "cipher": "aes128",
            "hash": "sha512",
            "model": "openssl",
        }
        modified_totem_options = {
            opt: ("no" if opt == "block_unlisted_ips" else val + "0")
            for opt, val in TOTEM_OPTIONS.items()
        }
        self.config.env.push_corosync_conf(
            corosync_conf_text=fixture_totem(
                transport_options=modified_transport_options,
                compression_options=modified_compression_options,
                crypto_options=modified_crypto_options,
                totem_options=modified_totem_options,
            ),
            need_stopped_cluster=need_stopped_cluster,
        )
        cluster.config_update(
            self.env_assist.get_env(),
            modified_transport_options,
            modified_compression_options,
            modified_crypto_options,
            modified_totem_options,
        )
        self.env_assist.assert_reports([])

    def test_modify_all_options(self):
        self._test_modify(
            {
                "ip_version": "ipv4-6",
                "link_mode": "passive",
                "knet_pmtud_interval": "1000",
            },
            need_stopped_cluster=True,
        )

    def test_modify_no_need_for_stopped_cluster(self):
        self._test_modify(
            {"link_mode": "passive", "knet_pmtud_interval": "1000"},
            need_stopped_cluster=False,
        )

    def test_remove_all_options(self):
        self.config.corosync_conf.load_content(
            fixture_totem(
                transport_options=TRANSPORT_OPTIONS,
                compression_options=COMPRESSION_OPTIONS,
                crypto_options=CRYPTO_OPTIONS,
                totem_options=TOTEM_OPTIONS,
            )
        )
        self.config.env.push_corosync_conf(
            corosync_conf_text=fixture_totem(),
            need_stopped_cluster=True,
        )
        cluster.config_update(
            self.env_assist.get_env(),
            {option: "" for option in ALLOWED_KNET_TRANSPORT_OPTIONS},
            {option: "" for option in ALLOWED_COMPRESSION_OPTIONS},
            {option: "" for option in ALLOWED_CRYPTO_OPTIONS},
            {option: "" for option in ALLOWED_TOTEM_OPTIONS},
        )
        self.env_assist.assert_reports([])

    def test_unknown_options_and_values(self):
        self.config.corosync_conf.load_content(fixture_totem())
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update(
                self.env_assist.get_env(),
                {"unknown": "val", "ip_version": "4"},
                {"level": "high", "unknown": "val"},
                {"cipher": "strong", "unknown": "val"},
                {"unknown": "val", "downcheck": "check"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="totem",
                    allowed=ALLOWED_TOTEM_OPTIONS,
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="downcheck",
                    option_value="check",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="knet transport",
                    allowed=ALLOWED_KNET_TRANSPORT_OPTIONS,
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="ip_version",
                    option_value="4",
                    allowed_values=("ipv4", "ipv6", "ipv4-6", "ipv6-4"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="compression",
                    allowed=ALLOWED_COMPRESSION_OPTIONS,
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="level",
                    option_value="high",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="crypto",
                    allowed=ALLOWED_CRYPTO_OPTIONS,
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="cipher",
                    option_value="strong",
                    allowed_values=("none", "aes256", "aes192", "aes128"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_MUST_BE_ENABLED_AS_WELL,
                    option_name="cipher",
                    prerequisite_name="hash",
                    option_type="crypto",
                    prerequisite_type="crypto",
                ),
            ]
        )

    def test_unsupported_transport(self):
        self.config.corosync_conf.load_content(
            fixture_totem(transport_type="keynet")
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update(
                self.env_assist.get_env(),
                {"unknown": "val", "ip_version": "4"},
                {"level": "high", "unknown": "val"},
                {"cipher": "strong", "unknown": "val"},
                {"unknown": "val", "downcheck": "check"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="totem",
                    allowed=ALLOWED_TOTEM_OPTIONS,
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.COROSYNC_CONFIG_UNSUPPORTED_TRANSPORT,
                    actual_transport="keynet",
                    supported_transport_types=["knet", "udp", "udpu"],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="downcheck",
                    option_value="check",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_udp_transport_unsupported_options(self):
        self.config.corosync_conf.load_content(
            fixture_totem(transport_type="udp")
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update(
                self.env_assist.get_env(),
                {"knet_pmtud_interval": "interval"},
                {"level": "high", "unknown": "val"},
                {"cipher": "strong", "unknown": "val"},
                {"downcheck": "check"},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="downcheck",
                    option_value="check",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["knet_pmtud_interval"],
                    option_type="udp/udpu transport",
                    allowed=["ip_version", "netmtu"],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.COROSYNC_TRANSPORT_UNSUPPORTED_OPTIONS,
                    option_type="compression",
                    actual_transport="udp/udpu",
                    required_transports=["knet"],
                ),
                fixture.error(
                    report_codes.COROSYNC_TRANSPORT_UNSUPPORTED_OPTIONS,
                    option_type="crypto",
                    actual_transport="udp/udpu",
                    required_transports=["knet"],
                ),
            ]
        )


class UpdateConfigLocal(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_empty_config_no_change(self):
        self.assertEqual(
            cluster.config_update_local(
                self.env_assist.get_env(), b"", {}, {}, {}, {}
            ),
            b"",
        )

    def test_invalid_custom_corosync_conf(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update_local(
                self.env_assist.get_env(),
                b"totem {\n    option.name: value\n}",
                {},
                {},
                {},
                {},
            ),
            [
                fixture.error(
                    report_codes.COROSYNC_CONFIG_CANNOT_SAVE_INVALID_NAMES_VALUES,
                    section_name_list=[],
                    attribute_name_list=["totem.option.name"],
                    attribute_value_pairs=[],
                )
            ],
            expected_in_processor=False,
        )

    def test_corosyncconf_parse_error(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update_local(
                self.env_assist.get_env(),
                b"this is not\na valid corosync.conf file\n",
                {},
                {},
                {},
                {},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE
                )
            ],
        )

    def test_add_modify_remove_options(self):
        before = fixture_totem(
            transport_options={
                "ip_version": "ipv4",
                "knet_pmtud_interval": "10",
            },
            compression_options={"level": "9", "threshold": "100"},
            crypto_options={"cipher": "none", "model": "openssl"},
            totem_options={"token": "3000", "join": "50"},
        )
        after = dedent(
            """\
            totem {
                transport: knet
                token: 1000
                ip_version: ipv6
                knet_compression_level: 5
                crypto_cipher: aes128
                downcheck: 5
                link_mode: active
                knet_compression_model: zlib
                crypto_hash: sha256
            }
            """
        )
        self.assertEqual(
            cluster.config_update_local(
                self.env_assist.get_env(),
                before.encode(),
                {
                    "ip_version": "ipv6",
                    "link_mode": "active",
                    "knet_pmtud_interval": "",
                },
                {"level": "5", "model": "zlib", "threshold": ""},
                {"cipher": "aes128", "hash": "sha256", "model": ""},
                {"token": "1000", "downcheck": "5", "join": ""},
            ),
            after.encode(),
        )

    def test_some_validator_errors(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster.config_update_local(
                self.env_assist.get_env(),
                fixture_totem(transport_type="udp").encode(),
                {"knet_pmtud_interval": "100"},
                {},
                {"cipher": "none", "hash": "none"},
                {"token": "notanumber"},
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="token",
                    option_value="notanumber",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["knet_pmtud_interval"],
                    option_type="udp/udpu transport",
                    allowed=["ip_version", "netmtu"],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.COROSYNC_TRANSPORT_UNSUPPORTED_OPTIONS,
                    option_type="crypto",
                    actual_transport="udp/udpu",
                    required_transports=["knet"],
                ),
            ]
        )



from unittest import TestCase

from pcs_test.tools.misc import outdent


class OutdentTest(TestCase):
    def test_returns_the_same_text_when_not_indented(self):
        text = "\n".join(
            [
                "first line",
                "  second line",
                "    third line",
            ]
        )
        self.assertEqual(text, outdent(text))

    def test_remove_the_smallest_indentation(self):
        self.assertEqual(
            "\n".join(
                [
                    "  first line",
                    "second line",
                    "  third line",
                ]
            ),
            outdent(
                "\n".join(
                    [
                        "    first line",
                        "  second line",
                        "    third line",
                    ]
                )
            ),
        )

    def test_very_ugly_indented_text(self):
        self.assertEqual(
            """\
Cluster Name: test99
  Options:
""",
            outdent(
                """\
                Cluster Name: test99
                  Options:
                """
            ),
        )



from unittest import mock, TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from pcs import settings
from pcs.common.reports import (
    codes as report_codes,
    messages as report_messages,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.errors import LibraryError
from pcs.lib.commands import resource

# This class does not focusing on validation testing, there are validator tests
# for that in pcs_test.tier0.lib.cib.test_resource_relation
class GroupAdd(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_existing_group(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="R1" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G", ["R3", "R1"])

    def test_new_group(self):
        resources_before = """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <primitive id="R2" />
                <group id="G">
                    <primitive id="R3" />
                    <primitive id="R1" />
                </group>
            </resources>
        """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G", ["R3", "R1"])

    def test_move_from_existing_to_new(self):
        resources_before = """
                    <resources>
                        <group id="G1">
                            <primitive id="RG1" />
                            <primitive id="RG2" />
                            <primitive id="RG3" />
                        </group>
                        <primitive id="R4" />
                    </resources>
                """
        resources_after = """
                    <resources>
                        <group id="G1">
                            <primitive id="RG1" />
                            <primitive id="RG2" />
                        </group>
                        <group id="G2">
                            <primitive id="RG3" />
                            <primitive id="R4" />
                        </group>
                    </resources>
                """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G2", ["RG3", "R4"])

    def test_move_from_existing_to_existing(self):
        resources_before = """
                    <resources>
                        <group id="G">
                            <primitive id="RG1" />
                        </group>
                        <primitive id="R1" />
                        <primitive id="R2" />
                        <primitive id="R3" />
                    </resources>
                """
        resources_after = """
                    <resources>
                        <group id="G">
                            <primitive id="RG1" />
                            <primitive id="R3" />
                            <primitive id="R1" />
                            <primitive id="R2" />
                        </group>
                    </resources>
                """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G", ["R3", "R1", "R2"])

    def _assert_with_adjacent(self, adjacent_id, after_adjacent):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="RG2" />
                    <primitive id="RGX" />
                </group>
                <primitive id="R1" />
                <primitive id="R2" />
                <primitive id="R3" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                    <primitive id="R3" />
                    <primitive id="RGX" />
                    <primitive id="R1" />
                    <primitive id="RG2" />
                </group>
                <primitive id="R2" />
            </resources>
        """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R3", "RGX", "R1"],
            adjacent_resource_id=adjacent_id,
            put_after_adjacent=after_adjacent,
        )

    def test_after_adjacent(self):
        self._assert_with_adjacent("RG1", True)

    def test_before_adjacent(self):
        self._assert_with_adjacent("RG2", False)

    def test_remove_empty_group(self):
        resources_before = """
            <resources>
                
                <group id="X">
                    <primitive id="RX1" />
                </group>
                <group id="Y">
                    <primitive id="RY1" />
                    <primitive id="RY2" />
                </group>
            </resources>
        """
        resources_after = """
            <resources>
                <group id="Y">
                    <primitive id="RY1" />
                </group>
                <group id="G">
                    <primitive id="RX1" />
                    <primitive id="RY2" />
                </group>
            </resources>
        """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G", ["RX1", "RY2"])

    def test_remove_empty_group_fail(self):
        constraints_before = """
            <constraints>
                <rsc_order first="X" first-action="start" 
                id="order-X-R1-mandatory" then="R1" 
                then-action="start"/>
            </constraints>
        """
        resources_before = """
            <resources>
                <primitive id="R1" />
                <group id="X">
                    <primitive id="RX1" />
                </group>
                <group id="Y">
                    <primitive id="RY1" />
                    <primitive id="RY2" />
                </group>
            </resources>
        """
        resources_after = """
            <resources>
                <primitive id="R1" />
                <group id="Y">
                    <primitive id="RY1" />
                </group>
                <group id="G">
                    <primitive id="RX1" />
                    <primitive id="RY2" />
                </group>
            </resources>
        """

        self.config.runner.cib.load(
            resources=resources_before, constraints=constraints_before
        ).env.push_cib(
            exception=LibraryError(
                ReportItem.error(
                    report_messages.CibPushError("stderr", "stdout")
                )
            ),
            resources=resources_after,
            constraints=constraints_before,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "G",
                ["RX1", "RY2"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CANNOT_LEAVE_GROUP_EMPTY_AFTER_MOVE,
                    group_id="X",
                    inner_resource_ids=["RX1"],
                ),
            ]
        )

    def test_remove_empty_clone(self):
        resources_before = """
            <resources>
                <clone id="RC1-group-clone">
                    <group id="RC1-group">
                        <primitive id="RC1" />
                    </group>
                </clone>
                <master id="RM1-group-master">
                    <group id="RM1-group">
                        <primitive id="RM1" />
                    </group>
                </master>
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="RC1" />
                    <primitive id="RM1" />
                </group>
            </resources>
        """
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after)
        )
        resource.group_add(self.env_assist.get_env(), "G", ["RC1", "RM1"])

    def test_remove_empty_clone_fail(self):
        constraints_before = """
            <constraints>
                <rsc_order first="RM1-group-master" first-action="start" 
                id="order-RM1-group-master-R1-mandatory" then="R1" 
                then-action="start"/>
                <rsc_order first="RC1-group-clone" first-action="start" 
                id="order-RC1-group-clone-R1-mandatory" then="R2" 
                then-action="start"/>
            </constraints>
        """
        resources_before = """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
                <clone id="RC1-group-clone">
                    <group id="RC1-group">
                        <primitive id="RC1" />
                    </group>
                </clone>
                <master id="RM1-group-master">
                    <group id="RM1-group">
                        <primitive id="RM1" />
                    </group>
                </master>
            </resources>
        """
        resources_after = """
            <resources>
                <primitive id="R1"></primitive>
                <primitive id="R2"></primitive>
                <group id="G">
                    <primitive id="RC1"></primitive>
                    <primitive id="RM1"></primitive>
                </group>
            </resources>
        """

        self.config.runner.cib.load(
            resources=resources_before, constraints=constraints_before
        ).env.push_cib(
            exception=LibraryError(
                ReportItem.error(
                    report_messages.CibPushError("stderr", "stdout")
                )
            ),
            resources=resources_after,
            constraints=constraints_before,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "G",
                ["RC1", "RM1"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CANNOT_LEAVE_GROUP_EMPTY_AFTER_MOVE,
                    group_id="RC1-group",
                    inner_resource_ids=["RC1"],
                ),
                fixture.error(
                    report_codes.CANNOT_LEAVE_GROUP_EMPTY_AFTER_MOVE,
                    group_id="RM1-group",
                    inner_resource_ids=["RM1"],
                ),
            ]
        )

    def test_validation(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <primitive id="R1">
                    <meta_attributes id="R1-meta_attributes" />
                </primitive>
                <primitive id="R2">
                    <meta_attributes id="R2-meta_attributes" />
                </primitive>
                <clone id="RC1-clone">
                    <primitive id="RC1" />
                </clone>
            </resources>
        """
        (self.config.runner.cib.load(resources=resources_before))
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "R1-meta_attributes",
                ["R2", "R4", "R3", "R2-meta_attributes", "RC1-clone", "RC1"],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-meta_attributes",
                    expected_types=["group"],
                    current_type="meta_attributes",
                ),
                fixture.report_not_found("R4", context_type="resources"),
                fixture.report_not_found("R3", context_type="resources"),
                fixture.error(
                    report_codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R2-meta_attributes",
                    expected_types=[
                        "bundle",
                        "clone",
                        "group",
                        "master",
                        "primitive",
                    ],
                    current_type="meta_attributes",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1-clone",
                    resource_type="clone",
                ),
                fixture.error(
                    report_codes.CANNOT_GROUP_RESOURCE_WRONG_TYPE,
                    resource_id="RC1",
                    resource_type="clone",
                ),
            ]
        )

    def test_validation_adjacent(self):
        resources_before = """
            <resources>
                <group id="G">
                    <primitive id="RG1" />
                </group>
                <group id="X">
                    <primitive id="RX1" />
                </group>
                <primitive id="R1" />
            </resources>
        """
        (self.config.runner.cib.load(resources=resources_before))
        self.env_assist.assert_raise_library_error(
            lambda: resource.group_add(
                self.env_assist.get_env(),
                "G",
                ["R1"],
                adjacent_resource_id="RX1",
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    report_codes.CANNOT_GROUP_RESOURCE_ADJACENT_RESOURCE_NOT_IN_GROUP,
                    adjacent_resource_id="RX1",
                    group_id="G",
                ),
            ]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class GroupAddWait(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        resources_before = """
            <resources>
                <primitive id="R1" />
                <primitive id="R2" />
            </resources>
        """
        resources_after = """
            <resources>
                <group id="G">
                    <primitive id="R1" />
                    <primitive id="R2" />
                </group>
            </resources>
        """
        self.timeout = 10
        (
            self.config.runner.cib.load(
                resources=resources_before
            ).env.push_cib(resources=resources_after, wait=self.timeout)
        )

    def test_group_running(self):
        resources_state = """
            <resources>
                <group id="G" number_resources="2">
                    <resource id="R1" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                    <resource id="R2" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                </group>
            </resources>
        """
        self.config.runner.pcmk.load_state(resources=resources_state)
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R1", "R2"],
            wait=self.timeout,
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_running("G", {"Started": ["node1"]}),
            ]
        )

    def test_group_not_running(self):
        resources_state = """
            <resources>
                <group id="G" number_resources="2">
                    <resource id="R1" role="Started" nodes_running_on="1">
                         <node name="node1" id="1" cached="false" />
                     </resource>
                    <resource id="R2" role="Stopped" nodes_running_on="0">
                     </resource>
                </group>
            </resources>
        """
        self.config.runner.pcmk.load_state(resources=resources_state)
        resource.group_add(
            self.env_assist.get_env(),
            "G",
            ["R1", "R2"],
            wait=self.timeout,
        )
        self.env_assist.assert_reports(
            [
                fixture.report_resource_not_running("G"),
            ]
        )



import logging
from contextlib import contextmanager
from unittest import mock, TestCase
from functools import partial
from lxml import etree

from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import create_patcher

from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes
from pcs.lib.cib.tools import IdProvider
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError

from pcs.lib.commands import node as lib

# pylint: disable=protected-access

mocked_cib = etree.fromstring("<cib />")

patch_env = partial(mock.patch.object, LibraryEnvironment)
patch_command = create_patcher("pcs.lib.commands.node")

create_env = partial(
    LibraryEnvironment,
    mock.MagicMock(logging.Logger),
    MockLibraryReportProcessor(),
)


def fixture_node(order_num):
    node = mock.MagicMock(attrs=mock.MagicMock())
    node.attrs.name = "node-{0}".format(order_num)
    return node


class StandbyMaintenancePassParameters(TestCase):
    def setUp(self):
        self.lib_env = "lib_env"
        self.nodes = "nodes"
        self.wait = "wait"
        self.standby_on = {"standby": "on"}
        self.standby_off = {"standby": ""}
        self.maintenance_on = {"maintenance": "on"}
        self.maintenance_off = {"maintenance": ""}


@patch_command("_set_instance_attrs_local_node")
class StandbyMaintenancePassParametersLocal(StandbyMaintenancePassParameters):
    def test_standby(self, mock_doer):
        lib.standby_unstandby_local(self.lib_env, True, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_on, self.wait
        )

    def test_unstandby(self, mock_doer):
        lib.standby_unstandby_local(self.lib_env, False, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_off, self.wait
        )

    def test_maintenance(self, mock_doer):
        lib.maintenance_unmaintenance_local(self.lib_env, True, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_on, self.wait
        )

    def test_unmaintenance(self, mock_doer):
        lib.maintenance_unmaintenance_local(self.lib_env, False, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_off, self.wait
        )


@patch_command("_set_instance_attrs_node_list")
class StandbyMaintenancePassParametersList(StandbyMaintenancePassParameters):
    def test_standby(self, mock_doer):
        lib.standby_unstandby_list(self.lib_env, True, self.nodes, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_on, self.nodes, self.wait
        )

    def test_unstandby(self, mock_doer):
        lib.standby_unstandby_list(self.lib_env, False, self.nodes, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_off, self.nodes, self.wait
        )

    def test_maintenance(self, mock_doer):
        lib.maintenance_unmaintenance_list(
            self.lib_env, True, self.nodes, self.wait
        )
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_on, self.nodes, self.wait
        )

    def test_unmaintenance(self, mock_doer):
        lib.maintenance_unmaintenance_list(
            self.lib_env, False, self.nodes, self.wait
        )
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_off, self.nodes, self.wait
        )


@patch_command("_set_instance_attrs_all_nodes")
class StandbyMaintenancePassParametersAll(StandbyMaintenancePassParameters):
    def test_standby(self, mock_doer):
        lib.standby_unstandby_all(self.lib_env, True, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_on, self.wait
        )

    def test_unstandby(self, mock_doer):
        lib.standby_unstandby_all(self.lib_env, False, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.standby_off, self.wait
        )

    def test_maintenance(self, mock_doer):
        lib.maintenance_unmaintenance_all(self.lib_env, True, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_on, self.wait
        )

    def test_unmaintenance(self, mock_doer):
        lib.maintenance_unmaintenance_all(self.lib_env, False, self.wait)
        mock_doer.assert_called_once_with(
            self.lib_env, self.maintenance_off, self.wait
        )


class SetInstaceAttrsBase(TestCase):
    node_count = 2

    def setUp(self):
        self.cluster_nodes = [fixture_node(i) for i in range(self.node_count)]
        self.cib = etree.fromstring("<cib />")

        self.launch = {"pre": False, "post": False}

        @contextmanager
        def cib_runner_nodes_contextmanager(env, wait):
            # pylint: disable=unused-argument
            self.launch["pre"] = True
            yield (self.cib, "mock_runner", self.cluster_nodes)
            self.launch["post"] = True

        patcher = patch_command("cib_runner_nodes")
        self.addCleanup(patcher.stop)
        patcher.start().side_effect = cib_runner_nodes_contextmanager

    def assert_context_manager_launched(self, pre=False, post=False):
        self.assertEqual(self.launch, {"pre": pre, "post": post})

    def assert_call_with_provider(self, call, cib, node, attrs, nodes):
        self.assertEqual(call[0][0], cib)
        self.assertTrue(isinstance(call[0][1], IdProvider))
        self.assertEqual(call[0][2], node)
        self.assertEqual(call[0][3], attrs)
        self.assertEqual(call[1]["state_nodes"], nodes)


@patch_command("update_node_instance_attrs")
@patch_command("get_local_node_name")
class SetInstaceAttrsLocal(SetInstaceAttrsBase):
    node_count = 2

    def test_not_possible_with_cib_file(self, mock_name, mock_attrs):
        assert_raise_library_error(
            lambda: lib._set_instance_attrs_local_node(
                create_env(cib_data="<cib />"), "attrs", "wait"
            ),
            (
                severity.ERROR,
                report_codes.LIVE_ENVIRONMENT_REQUIRED_FOR_LOCAL_NODE,
                {},
            ),
        )
        self.assert_context_manager_launched(pre=False, post=False)
        mock_name.assert_not_called()
        mock_attrs.assert_not_called()

    def test_success(self, mock_name, mock_attrs):
        mock_name.return_value = "node-1"

        lib._set_instance_attrs_local_node(create_env(), "attrs", False)

        self.assert_context_manager_launched(pre=True, post=True)
        mock_name.assert_called_once_with("mock_runner")
        mock_attrs_calls = mock_attrs.call_args_list
        self.assertEqual(len(mock_attrs_calls), 1)
        self.assert_call_with_provider(
            mock_attrs_calls[0], self.cib, "node-1", "attrs", self.cluster_nodes
        )


@patch_command("update_node_instance_attrs")
class SetInstaceAttrsAll(SetInstaceAttrsBase):
    node_count = 2

    def test_success(self, mock_attrs):
        lib._set_instance_attrs_all_nodes(create_env(), "attrs", False)

        mock_attrs_calls = mock_attrs.call_args_list
        self.assertEqual(2, len(mock_attrs_calls))
        self.assert_call_with_provider(
            mock_attrs_calls[0], self.cib, "node-0", "attrs", self.cluster_nodes
        )
        self.assert_call_with_provider(
            mock_attrs_calls[1], self.cib, "node-1", "attrs", self.cluster_nodes
        )


@patch_command("update_node_instance_attrs")
class SetInstaceAttrsList(SetInstaceAttrsBase):
    node_count = 4

    def test_success(self, mock_attrs):
        lib._set_instance_attrs_node_list(
            create_env(), "attrs", ["node-1", "node-2"], False
        )

        self.assert_context_manager_launched(pre=True, post=True)
        mock_attrs_calls = mock_attrs.call_args_list
        self.assertEqual(2, len(mock_attrs_calls))
        self.assert_call_with_provider(
            mock_attrs_calls[0], self.cib, "node-1", "attrs", self.cluster_nodes
        )
        self.assert_call_with_provider(
            mock_attrs_calls[1], self.cib, "node-2", "attrs", self.cluster_nodes
        )

    def test_bad_node(self, mock_attrs):
        # pylint: disable=no-self-use
        assert_raise_library_error(
            lambda: lib._set_instance_attrs_node_list(
                create_env(), "attrs", ["node-1", "node-9"], False
            ),
            (
                severity.ERROR,
                report_codes.NODE_NOT_FOUND,
                {
                    "node": "node-9",
                    "searched_types": [],
                },
            ),
        )
        mock_attrs.assert_not_called()


@patch_env("push_cib")
class CibRunnerNodes(TestCase):
    def setUp(self):
        self.env = create_env()

    @patch_env("get_cib", lambda self: "mocked cib")
    @patch_env("cmd_runner", lambda self: "mocked cmd_runner")
    @patch_env("ensure_wait_satisfiable")
    @patch_command("ClusterState")
    @patch_env("get_cluster_state")
    def test_wire_together_all_expected_dependecies(
        self,
        get_cluster_state,
        cluster_state,
        ensure_wait_satisfiable,
        push_cib,
    ):
        cluster_state.return_value = mock.MagicMock(
            node_section=mock.MagicMock(nodes="nodes")
        )
        get_cluster_state.return_value = "mock get_cluster_state"
        wait = 10
        ensure_wait_satisfiable.return_value = wait

        with lib.cib_runner_nodes(self.env, wait) as (cib, runner, nodes):
            self.assertEqual(cib, "mocked cib")
            self.assertEqual(runner, "mocked cmd_runner")
            self.assertEqual(nodes, "nodes")
            ensure_wait_satisfiable.assert_called_once_with(wait)
            get_cluster_state.assert_called_once_with()
            cluster_state.assert_called_once_with("mock get_cluster_state")

        push_cib.assert_called_once_with(wait_timeout=wait)

    @patch_env("ensure_wait_satisfiable", mock.Mock(side_effect=LibraryError))
    def test_raises_when_wait_is_not_satisfiable(self, push_cib):
        def run():
            # pylint: disable=unused-variable
            with lib.cib_runner_nodes(self.env, "wait") as (cib, runner, nodes):
                pass

        self.assertRaises(LibraryError, run)
        push_cib.assert_not_called()



from unittest import TestCase

from lxml import etree

from pcs.common import reports
from pcs.lib.cib import stonith

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


class IsStonithEnabled(TestCase):
    def test_not_set(self):
        crm_config = etree.fromstring("<crm_config />")
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_enabled(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertTrue(stonith.is_stonith_enabled(crm_config))

    def test_set_to_disabled(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="abc" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_values(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))

    def test_multiple_sections(self):
        crm_config = etree.fromstring(
            """
            <crm_config>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                    <nvpair name="stonith-enabled" value="true" />
                </cluster_property_set>
                <cluster_property_set>
                    <nvpair name="stonith-enabled" value="false" />
                </cluster_property_set>
            </crm_config>
        """
        )
        self.assertFalse(stonith.is_stonith_enabled(crm_config))


class GetMisconfiguredResources(TestCase):
    def test_no_stonith(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="R" class="ocf" provider="pacemaker" type="Dummy">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        self.assertEqual(
            stonith.get_misconfigured_resources(resources), ([], [], [])
        )

    def test_all_ok(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                resources.findall("primitive[@id='S1']"),
                [],
                [],
            ),
        )

    def test_issues(self):
        resources = etree.fromstring(
            """
            <resources>
                <primitive id="S1" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="name" value="value" />
                        <nvpair name="method" value="onoff" />
                    </instance_attributes>
                </primitive>
                <primitive id="S2" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                    </instance_attributes>
                </primitive>
                <primitive id="S3" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
                <primitive id="S4" class="stonith" type="fence_something">
                    <instance_attributes>
                        <nvpair name="action" value="value" />
                        <nvpair name="method" value="cycle" />
                    </instance_attributes>
                </primitive>
            </resources>
        """
        )
        stonith1 = resources.find("primitive[@id='S1']")
        stonith2 = resources.find("primitive[@id='S2']")
        stonith3 = resources.find("primitive[@id='S3']")
        stonith4 = resources.find("primitive[@id='S4']")
        self.assertEqual(
            stonith.get_misconfigured_resources(resources),
            (
                [stonith1, stonith2, stonith3, stonith4],
                [stonith2, stonith4],
                [stonith3, stonith4],
            ),
        )


class ValidateStonithRestartlessUpdate(TestCase):
    RESOURCES = etree.fromstring(
        """
        <resources>
            <primitive id="supported" class="stonith" type="fence_scsi">
                <instance_attributes>
                    <nvpair name="devices" value="/dev/sda" />
                </instance_attributes>
            </primitive>
            <primitive id="empty" class="stonith" type="fence_scsi">
                <instance_attributes>
                    <nvpair id="empty-instance_attributes-devices"
                        name="devices" value="" />
                </instance_attributes>
            </primitive>
            <primitive id="no-devices" class="stonith" type="fence_scsi"/>
            <primitive id="unsupported_provider"
                class="stonith" provider="provider" type="fence_scsi"/>
            <primitive id="unsupported_type" class="stonith" type="fence_xvm"/>
            <primitive class="ocf" id="cp-01" provider="pacemaker" type="Dummy"/>
        </resources>
        """
    )

    def assert_unsupported_stonith_agent(self, resource_id, resource_type):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, resource_id
        )
        self.assertEqual(
            stonith_el,
            self.RESOURCES.find(f".//primitive[@id='{resource_id}']"),
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNSUPPORTED_AGENT,
                    resource_id=resource_id,
                    resource_type=resource_type,
                    supported_stonith_types=["fence_scsi"],
                )
            ],
        )

    def assert_no_devices(self, resource_id):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, resource_id
        )
        self.assertEqual(
            stonith_el,
            self.RESOURCES.find(f".//primitive[@id='{resource_id}']"),
        )
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM,
                    reason=(
                        "no devices option configured for stonith device "
                        f"'{resource_id}'"
                    ),
                    reason_type="other",
                )
            ],
        )

    def test_supported(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "supported"
        )
        self.assertEqual(
            stonith_el, self.RESOURCES.find(".//primitive[@id='supported']")
        )
        assert_report_item_list_equal(report_list, [])

    def test_nonexistent_id(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "non-existent"
        )
        self.assertEqual(stonith_el, None)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="non-existent",
                    expected_types=["primitive"],
                    context_type="resources",
                    context_id="",
                )
            ],
        )

    def test_not_a_resource_id(self):
        stonith_el, report_list = stonith.validate_stonith_restartless_update(
            self.RESOURCES, "empty-instance_attributes-devices"
        )
        self.assertEqual(stonith_el, None)
        assert_report_item_list_equal(
            report_list,
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="empty-instance_attributes-devices",
                    expected_types=["primitive"],
                    current_type="nvpair",
                )
            ],
        )

    def test_devices_empty(self):
        self.assert_no_devices("empty")

    def test_missing_devices_attr(self):
        self.assert_no_devices("no-devices")

    def test_unsupported_class(self):
        self.assert_unsupported_stonith_agent("cp-01", "Dummy")

    def test_unsupported_provider(self):
        self.assert_unsupported_stonith_agent(
            "unsupported_provider", "fence_scsi"
        )

    def test_unsupported_type(self):
        self.assert_unsupported_stonith_agent("unsupported_type", "fence_xvm")



# pylint: disable=too-many-lines
import os.path
from unittest import mock, TestCase
from lxml import etree

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
    assert_xml_equal,
    start_tag_error_text,
)
from pcs_test.tools import fixture, fixture_crm_mon
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.command_env.mock_runner import Call as RunnerCall
from pcs_test.tools.custom_mock import (
    TmpFileMock,
    TmpFileCall,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.xml import etree_to_str, XmlManipulation

from pcs import settings
from pcs.common.reports import ReportItemSeverity as Severity
from pcs.common.reports import codes as report_codes
from pcs.common.tools import Version
from pcs.common.types import CibRuleInEffectStatus
from pcs.lib.pacemaker import api_result
import pcs.lib.pacemaker.live as lib
from pcs.lib.external import CommandRunner

_EXITCODE_NOT_CONNECTED = 102


def get_runner(stdout="", stderr="", returncode=0, env_vars=None):
    runner = mock.MagicMock(spec_set=CommandRunner)
    runner.run.return_value = (stdout, stderr, returncode)
    runner.env_vars = env_vars if env_vars else {}
    return runner


def path(name):
    return os.path.join(settings.pacemaker_binaries, name)


class GetStatusFromApiResult(TestCase):
    def test_errors(self):
        self.assertEqual(
            lib.get_status_from_api_result(
                etree.fromstring(
                    fixture_crm_mon.error_xml(
                        123, "short message", ["error1", "error2"]
                    )
                )
            ),
            api_result.Status(123, "short message", ["error1", "error2"]),
        )

    def test_no_errors(self):
        self.assertEqual(
            lib.get_status_from_api_result(
                etree.fromstring(
                    fixture_crm_mon.error_xml(123, "short message")
                )
            ),
            api_result.Status(123, "short message", []),
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
@mock.patch.object(settings, "crm_mon_schema", rc("crm_mon_rng/crm_mon.rng"))
class GetClusterStatusXmlTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self._xml_summary = etree_to_str(
            etree.parse(rc("crm_mon.minimal.xml")).find("/summary")
        )

    def fixture_xml(self, transformed=False):
        as_xml = "--as-xml" if transformed else "--output-as xml"
        return f"""
            <pacemaker-result api-version="2.3" request="crm_mon {as_xml}">
              {self._xml_summary}
              <nodes />
              <resources />
              <status code="0" message="OK" />
            </pacemaker-result>
        """

    def fixture_xml_old(self):
        return f"""
            <crm_mon version="2.0.5">
              {self._xml_summary}
              <nodes />
              <resources />
            </crm_mon>
        """

    def place_call_old_format_help(self):
        self.config.calls.place(
            "local.crm_mon.help-all",
            RunnerCall(
                ["crm_mon", "--help-all"],
                stdout="this version only supports --as-xml option",
            ),
        )

    def place_call_old_format_xml(self, stdout="", stderr="", returncode=0):
        self.config.calls.place(
            "local.crm_mon.as-xml",
            RunnerCall(
                ["crm_mon", "--one-shot", "--inactive", "--as-xml"],
                stdout=stdout,
                stderr=stderr,
                returncode=returncode,
            ),
        )

    def test_success(self):
        self.config.runner.pcmk.load_state(stdout=self.fixture_xml())
        env = self.env_assist.get_env()
        assert_xml_equal(
            self.fixture_xml(),
            etree_to_str(lib.get_cluster_status_dom(env.cmd_runner())),
        )

    def test_error(self):
        self.config.runner.pcmk.load_state(
            stdout=fixture_crm_mon.error_xml(
                1, "an error", ["This is an error message"]
            ),
            returncode=1,
        )
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(
                report_codes.CRM_MON_ERROR,
                reason="an error\nThis is an error message",
            ),
        )

    def test_error_not_connected(self):
        self.config.runner.pcmk.load_state(
            stdout=fixture_crm_mon.error_xml_not_connected(),
            returncode=_EXITCODE_NOT_CONNECTED,
        )
        env = self.env_assist.get_env()
        with self.assertRaises(lib.PacemakerNotConnectedException) as cm:
            lib.get_cluster_status_dom(env.cmd_runner())
        assert_report_item_list_equal(
            cm.exception.args,
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason=(
                        "Not connected\n"
                        "crm_mon: Error: cluster is not available on this node"
                    ),
                ),
            ],
        )

    def test_error_not_xml(self):
        self.config.runner.pcmk.load_state(
            stdout="stdout text",
            stderr="stderr text",
            returncode=1,
        )
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(
                report_codes.CRM_MON_ERROR,
                reason="stderr text\nstdout text",
            ),
        )

    def test_invalid_xml(self):
        self.config.runner.pcmk.load_state(stdout="this is not an xml")
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(report_codes.BAD_CLUSTER_STATE_FORMAT),
        )

    def test_xml_doesnt_match_rng(self):
        self.config.runner.pcmk.load_state(stdout=self.fixture_xml_old())
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(report_codes.BAD_CLUSTER_STATE_FORMAT),
        )

    def test_success_old_format(self):
        self.place_call_old_format_help()
        self.place_call_old_format_xml(stdout=self.fixture_xml_old())
        env = self.env_assist.get_env()
        assert_xml_equal(
            self.fixture_xml(transformed=True),
            etree_to_str(lib.get_cluster_status_dom(env.cmd_runner())),
        )

    def test_error_old_format(self):
        self.place_call_old_format_help()
        self.place_call_old_format_xml(
            stdout="stdout text",
            stderr="stderr text",
            returncode=1,
        )
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(
                report_codes.CRM_MON_ERROR,
                reason="stderr text\nstdout text",
            ),
        )

    def test_invalid_xml_old_format(self):
        self.place_call_old_format_help()
        self.place_call_old_format_xml(stdout="this is not an xml")
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(report_codes.BAD_CLUSTER_STATE_FORMAT),
        )

    def test_xml_doesnt_match_rng_old_format(self):
        self.place_call_old_format_help()
        self.place_call_old_format_xml(stdout=self.fixture_xml())
        env = self.env_assist.get_env()
        assert_raise_library_error(
            lambda: lib.get_cluster_status_dom(env.cmd_runner()),
            fixture.error(report_codes.BAD_CLUSTER_STATE_FORMAT),
        )


class GetClusterStatusText(TestCase):
    def setUp(self):
        self.mock_fencehistory_supported = mock.patch(
            "pcs.lib.pacemaker.live.is_fence_history_supported_status",
            return_value=True,
        )
        self.mock_fencehistory_supported.start()
        self.expected_stdout = "cluster status"
        self.expected_stderr = ""
        self.expected_retval = 0

    def tearDown(self):
        self.mock_fencehistory_supported.stop()

    def get_runner(self, stdout=None, stderr=None, retval=None):
        return get_runner(
            self.expected_stdout if stdout is None else stdout,
            self.expected_stderr if stderr is None else stderr,
            self.expected_retval if retval is None else retval,
        )

    def test_success_minimal(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, False
        )

        mock_runner.run.assert_called_once_with(
            [settings.crm_mon, "--one-shot", "--inactive"]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_verbose(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with(
            [
                settings.crm_mon,
                "--one-shot",
                "--inactive",
                "--show-detail",
                "--show-node-attributes",
                "--failcounts",
                "--fence-history=3",
            ]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_no_fence_history(self):
        self.mock_fencehistory_supported.stop()
        self.mock_fencehistory_supported = mock.patch(
            "pcs.lib.pacemaker.live.is_fence_history_supported_status",
            return_value=False,
        )
        self.mock_fencehistory_supported.start()

        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with(
            [
                settings.crm_mon,
                "--one-shot",
                "--inactive",
                "--show-detail",
                "--show-node-attributes",
                "--failcounts",
            ]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_hide_inactive(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, True, False
        )

        mock_runner.run.assert_called_once_with(
            [settings.crm_mon, "--one-shot"]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_success_hide_inactive_verbose(self):
        mock_runner = self.get_runner()
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, True, True
        )

        mock_runner.run.assert_called_once_with(
            [
                settings.crm_mon,
                "--one-shot",
                "--show-detail",
                "--show-node-attributes",
                "--failcounts",
                "--fence-history=3",
            ]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, [])

    def test_error(self):
        mock_runner = self.get_runner("stdout", "stderr", 1)
        assert_raise_library_error(
            lambda: lib.get_cluster_status_text(mock_runner, False, False),
            (
                fixture.error(
                    report_codes.CRM_MON_ERROR, reason="stderr\nstdout"
                )
            ),
        )
        mock_runner.run.assert_called_once_with(
            [settings.crm_mon, "--one-shot", "--inactive"]
        )

    def test_warnings(self):
        mock_runner = self.get_runner(
            stderr="msgA\nDEBUG: msgB\nmsgC\nDEBUG: msgd\n"
        )
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, False
        )

        mock_runner.run.assert_called_once_with(
            [settings.crm_mon, "--one-shot", "--inactive"]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(warnings, ["msgA", "msgC"])

    def test_warnings_verbose(self):
        mock_runner = self.get_runner(
            stderr="msgA\nDEBUG: msgB\nmsgC\nDEBUG: msgd\n"
        )
        real_status, warnings = lib.get_cluster_status_text(
            mock_runner, False, True
        )

        mock_runner.run.assert_called_once_with(
            [
                settings.crm_mon,
                "--one-shot",
                "--inactive",
                "--show-detail",
                "--show-node-attributes",
                "--failcounts",
                "--fence-history=3",
            ]
        )
        self.assertEqual(self.expected_stdout, real_status)
        self.assertEqual(
            warnings, ["msgA", "DEBUG: msgB", "msgC", "DEBUG: msgd"]
        )


class GetCibXmlTest(TestCase):
    # pylint: disable=no-self-use
    def test_success(self):
        expected_stdout = "<xml />"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        real_xml = lib.get_cib_xml(mock_runner)

        mock_runner.run.assert_called_once_with(
            [path("cibadmin"), "--local", "--query"]
        )
        self.assertEqual(expected_stdout, real_xml)

    def test_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.get_cib_xml(mock_runner),
            (
                Severity.ERROR,
                report_codes.CIB_LOAD_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [path("cibadmin"), "--local", "--query"]
        )

    def test_success_scope(self):
        expected_stdout = "<xml />"
        expected_stderr = ""
        expected_retval = 0
        scope = "test_scope"
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        real_xml = lib.get_cib_xml(mock_runner, scope)

        mock_runner.run.assert_called_once_with(
            [
                path("cibadmin"),
                "--local",
                "--query",
                "--scope={0}".format(scope),
            ]
        )
        self.assertEqual(expected_stdout, real_xml)

    def test_scope_error(self):
        expected_stdout = "some info"
        # yes, the numbers do not match, tested and verified with
        # pacemaker-2.0.0-1.fc29.1.x86_64
        expected_stderr = (
            "Call cib_query failed (-6): No such device or address"
        )
        expected_retval = 105
        scope = "test_scope"
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.get_cib_xml(mock_runner, scope=scope),
            (
                Severity.ERROR,
                report_codes.CIB_LOAD_ERROR_SCOPE_MISSING,
                {
                    "scope": scope,
                    "reason": expected_stderr + "\n" + expected_stdout,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [
                path("cibadmin"),
                "--local",
                "--query",
                "--scope={0}".format(scope),
            ]
        )


class GetCibTest(TestCase):
    # pylint: disable=no-self-use
    def test_success(self):
        xml = "<xml />"
        assert_xml_equal(xml, str(XmlManipulation((lib.get_cib(xml)))))

    @mock.patch("pcs.lib.pacemaker.live.xml_fromstring")
    def test_invalid_xml(self, xml_fromstring_mock):
        reason = "custom reason"
        xml_fromstring_mock.side_effect = etree.XMLSyntaxError(reason, 1, 1, 1)
        xml = "<invalid><xml />"
        assert_raise_library_error(
            lambda: lib.get_cib(xml),
            fixture.error(
                report_codes.CIB_LOAD_ERROR_BAD_FORMAT,
                reason=f"{reason} (line 1)",
            ),
        )
        xml_fromstring_mock.assert_called_once_with(xml)


class Verify(TestCase):
    def test_run_on_live_cib(self):
        runner = get_runner()
        self.assertEqual(lib.verify(runner), ("", "", 0, False))
        runner.run.assert_called_once_with(
            [path("crm_verify"), "--live-check"],
        )

    def test_run_on_mocked_cib(self):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})

        self.assertEqual(lib.verify(runner), ("", "", 0, False))
        runner.run.assert_called_once_with(
            [path("crm_verify"), "--xml-file", fake_tmp_file],
        )

    def test_run_verbose(self):
        runner = get_runner()
        self.assertEqual(lib.verify(runner, verbose=True), ("", "", 0, False))
        runner.run.assert_called_once_with(
            [path("crm_verify"), "-V", "-V", "--live-check"],
        )

    def test_run_verbose_on_mocked_cib(self):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})

        self.assertEqual(lib.verify(runner, verbose=True), ("", "", 0, False))
        runner.run.assert_called_once_with(
            [path("crm_verify"), "-V", "-V", "--xml-file", fake_tmp_file],
        )

    @staticmethod
    def get_in_out_filtered_stderr():
        in_stderr = (
            (
                "Errors found during check: config not valid\n",
                "  -V may provide more details\n",
            ),
            (
                "Warnings found during check: config may not be valid\n",
                "  Use -V -V for more detail\n",
            ),
            (
                "some output\n",
                "another output\n",
                "-V -V -V more details...\n",
            ),
            (
                "some output\n",
                "before-V -V -V in the middle more detailafter\n",
                "another output\n",
            ),
        )
        out_stderr = []
        for input_lines in in_stderr:
            out_stderr.append(
                [line for line in input_lines if "-V" not in line]
            )
        return zip(in_stderr, out_stderr)

    @staticmethod
    def get_in_out_unfiltered_data():
        in_out_data = (
            ("no verbose option in stderr\n",),
            (
                "some output\n",
                "Options '-V -V' do not match\n",
                "because line missing 'more details'\n",
            ),
        )
        return zip(in_out_data, in_out_data)

    def subtest_filter_stderr_and_can_be_more_verbose(
        self,
        in_out_tuple_list,
        can_be_more_verbose,
        verbose=False,
    ):
        fake_tmp_file = "/fake/tmp/file"
        runner = get_runner(env_vars={"CIB_file": fake_tmp_file})
        for in_stderr, out_stderr in in_out_tuple_list:
            with self.subTest(in_stderr=in_stderr, out_stderr=out_stderr):
                runner = get_runner(
                    stderr="".join(in_stderr),
                    returncode=78,
                    env_vars={"CIB_file": fake_tmp_file},
                )
                self.assertEqual(
                    lib.verify(runner, verbose=verbose),
                    ("", "".join(out_stderr), 78, can_be_more_verbose),
                )
                args = [path("crm_verify")]
                if verbose:
                    args.extend(["-V", "-V"])
                args.extend(["--xml-file", fake_tmp_file])
                runner.run.assert_called_once_with(args)

    def test_error_can_be_more_verbose(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            self.get_in_out_filtered_stderr(),
            True,
        )

    def test_error_cannot_be_more_verbose(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            self.get_in_out_unfiltered_data(),
            False,
        )

    def test_error_cannot_be_more_verbose_in_verbose_mode(self):
        self.subtest_filter_stderr_and_can_be_more_verbose(
            (
                list(self.get_in_out_filtered_stderr())
                + list(self.get_in_out_unfiltered_data())
            ),
            False,
            verbose=True,
        )


class ReplaceCibConfigurationTest(TestCase):
    # pylint: disable=no-self-use
    def test_success(self):
        xml = "<xml/>"
        expected_stdout = "expected output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        lib.replace_cib_configuration(
            mock_runner, XmlManipulation.from_str(xml).tree
        )

        mock_runner.run.assert_called_once_with(
            [
                path("cibadmin"),
                "--replace",
                "--verbose",
                "--xml-pipe",
                "--scope",
                "configuration",
            ],
            stdin_string=xml,
        )

    def test_error(self):
        xml = "<xml/>"
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.replace_cib_configuration(
                mock_runner, XmlManipulation.from_str(xml).tree
            ),
            (
                Severity.ERROR,
                report_codes.CIB_PUSH_ERROR,
                {
                    "reason": expected_stderr,
                    "pushed_cib": expected_stdout,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [
                path("cibadmin"),
                "--replace",
                "--verbose",
                "--xml-pipe",
                "--scope",
                "configuration",
            ],
            stdin_string=xml,
        )


class UpgradeCibTest(TestCase):
    # pylint: disable=protected-access
    # pylint: disable=no-self-use
    def test_success(self):
        mock_runner = get_runner("", "", 0)
        lib._upgrade_cib(mock_runner)
        mock_runner.run.assert_called_once_with(
            [settings.cibadmin, "--upgrade", "--force"]
        )

    def test_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )
        assert_raise_library_error(
            lambda: lib._upgrade_cib(mock_runner),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                },
            ),
        )
        mock_runner.run.assert_called_once_with(
            [settings.cibadmin, "--upgrade", "--force"]
        )


@mock.patch("pcs.lib.pacemaker.live.get_cib_xml")
@mock.patch("pcs.lib.pacemaker.live._upgrade_cib")
class EnsureCibVersionTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)
        self.cib = etree.XML('<cib validate-with="pacemaker-2.3.4"/>')

    def test_same_version(self, mock_upgrade, mock_get_cib):
        actual_cib, was_upgraded = lib.ensure_cib_version(
            self.mock_runner, self.cib, Version(2, 3, 4)
        )
        self.assertEqual(self.cib, actual_cib)
        self.assertFalse(was_upgraded)
        mock_upgrade.assert_not_called()
        mock_get_cib.assert_not_called()

    def test_higher_version(self, mock_upgrade, mock_get_cib):
        actual_cib, was_upgraded = lib.ensure_cib_version(
            self.mock_runner, self.cib, Version(2, 3, 3)
        )
        self.assertEqual(self.cib, actual_cib)
        self.assertFalse(was_upgraded)
        mock_upgrade.assert_not_called()
        mock_get_cib.assert_not_called()

    def test_upgraded_same_version(self, mock_upgrade, mock_get_cib):
        expected_cib = '<cib validate-with="pacemaker-2.3.5"/>'
        mock_get_cib.return_value = expected_cib
        actual_cib, was_upgraded = lib.ensure_cib_version(
            self.mock_runner, self.cib, Version(2, 3, 5)
        )
        assert_xml_equal(expected_cib, etree.tostring(actual_cib).decode())
        self.assertTrue(was_upgraded)
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_upgraded_higher_version(self, mock_upgrade, mock_get_cib):
        expected_cib = '<cib validate-with="pacemaker-2.3.6"/>'
        mock_get_cib.return_value = expected_cib
        actual_cib, was_upgraded = lib.ensure_cib_version(
            self.mock_runner, self.cib, Version(2, 3, 5)
        )
        assert_xml_equal(expected_cib, etree.tostring(actual_cib).decode())
        self.assertTrue(was_upgraded)
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_upgraded_lower_version(self, mock_upgrade, mock_get_cib):
        mock_get_cib.return_value = etree.tostring(self.cib).decode()
        assert_raise_library_error(
            lambda: lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 5)
            ),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION,
                {"required_version": "2.3.5", "current_version": "2.3.4"},
            ),
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_upgraded_lower_version_dont_fail(self, mock_upgrade, mock_get_cib):
        expected_cib = '<cib validate-with="pacemaker-2.3.4"/>'
        mock_get_cib.return_value = expected_cib
        actual_cib, was_upgraded = lib.ensure_cib_version(
            self.mock_runner,
            self.cib,
            Version(2, 3, 5),
            fail_if_version_not_met=False,
        )
        assert_xml_equal(expected_cib, etree.tostring(actual_cib).decode())
        self.assertFalse(was_upgraded)
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)

    def test_cib_parse_error(self, mock_upgrade, mock_get_cib):
        mock_get_cib.return_value = "not xml"
        assert_raise_library_error(
            lambda: lib.ensure_cib_version(
                self.mock_runner, self.cib, Version(2, 3, 5)
            ),
            (
                Severity.ERROR,
                report_codes.CIB_UPGRADE_FAILED,
                {
                    "reason": start_tag_error_text(),
                },
            ),
        )
        mock_upgrade.assert_called_once_with(self.mock_runner)
        mock_get_cib.assert_called_once_with(self.mock_runner)


class SimulateCibXml(TestCase):
    def setUp(self):
        tmp_file_patcher = mock.patch("pcs.lib.pacemaker.live.get_tmp_file")
        self.addCleanup(tmp_file_patcher.stop)
        self.tmp_file_mock_obj = TmpFileMock()
        self.addCleanup(self.tmp_file_mock_obj.assert_all_done)
        tmp_file_mock = tmp_file_patcher.start()
        tmp_file_mock.side_effect = (
            self.tmp_file_mock_obj.get_mock_side_effect()
        )

    # pylint: disable=no-self-use
    def test_success(self):
        orig_cib_data = "orig cib"
        cib_file_name = "new_cib_file.tmp"
        transitions_file_name = "transitions.tmp"
        new_cib_data = "new cib data"
        transitions_data = "transitions data"
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall(cib_file_name, new_content=new_cib_data),
                TmpFileCall(
                    transitions_file_name, new_content=transitions_data
                ),
            ]
        )

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        result = lib.simulate_cib_xml(mock_runner, orig_cib_data)
        self.assertEqual(result[0], expected_stdout)
        self.assertEqual(result[1], transitions_data)
        self.assertEqual(result[2], new_cib_data)

        mock_runner.run.assert_called_once_with(
            [
                path("crm_simulate"),
                "--simulate",
                "--save-output",
                cib_file_name,
                "--save-graph",
                transitions_file_name,
                "--xml-pipe",
            ],
            stdin_string=orig_cib_data,
        )

    def test_error_creating_cib(self):
        err_msg = "some error"
        self.tmp_file_mock_obj.set_calls(
            [TmpFileCall("a file", orig_content=OSError(1, err_msg))]
        )
        mock_runner = get_runner()
        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=err_msg,
            ),
        )
        mock_runner.run.assert_not_called()

    def test_error_creating_transitions(self):
        err_msg = "some error"
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall("cib_file"),
                TmpFileCall("transitions", orig_content=OSError(1, err_msg)),
            ]
        )
        mock_runner = get_runner()
        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=err_msg,
            ),
        )
        mock_runner.run.assert_not_called()

    def test_error_running_simulate(self):
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall("cib_file"),
                TmpFileCall("transitions_file"),
            ]
        )
        expected_stdout = "some stdout"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason="some error",
            ),
        )

    def test_error_reading_cib(self):
        err_msg = "some error"
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall("cib_file", new_content=OSError(1, err_msg)),
                TmpFileCall("transitions", new_content=""),
            ]
        )

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=err_msg,
            ),
        )

    def test_error_reading_transitions(self):
        err_msg = "some error"
        self.tmp_file_mock_obj.set_calls(
            [
                TmpFileCall("cib_file", new_content=""),
                TmpFileCall("transitions", new_content=OSError(1, err_msg)),
            ]
        )

        expected_stdout = "simulate output"
        expected_stderr = ""
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.simulate_cib_xml(mock_runner, "<cib />"),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=err_msg,
            ),
        )


@mock.patch("pcs.lib.pacemaker.live.simulate_cib_xml")
class SimulateCib(TestCase):
    def setUp(self):
        self.runner = "mock runner"
        self.cib_xml = "<cib/>"
        self.cib = etree.fromstring(self.cib_xml)
        self.simulate_output = "  some output  "
        self.transitions = "<transitions/>"
        self.new_cib = "<new-cib/>"

    def test_success(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output,
            self.transitions,
            self.new_cib,
        )
        result = lib.simulate_cib(self.runner, self.cib)
        self.assertEqual(result[0], "some output")
        assert_xml_equal(self.transitions, etree_to_str(result[1]))
        assert_xml_equal(self.new_cib, etree_to_str(result[2]))
        mock_simulate.assert_called_once_with(self.runner, self.cib_xml)

    def test_invalid_cib(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output,
            "bad transitions",
            self.new_cib,
        )
        assert_raise_library_error(
            lambda: lib.simulate_cib(self.runner, self.cib),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=(
                    "Start tag expected, '<' not found, line 1, column 1 "
                    "(<string>, line 1)"
                ),
            ),
        )

    def test_invalid_transitions(self, mock_simulate):
        mock_simulate.return_value = (
            self.simulate_output,
            self.transitions,
            "bad new cib",
        )
        assert_raise_library_error(
            lambda: lib.simulate_cib(self.runner, self.cib),
            fixture.error(
                report_codes.CIB_SIMULATE_ERROR,
                reason=(
                    "Start tag expected, '<' not found, line 1, column 1 "
                    "(<string>, line 1)"
                ),
            ),
        )


class GetLocalNodeName(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        expected_name = "node-name"
        self.config.runner.pcmk.local_node_name(node_name=expected_name)
        env = self.env_assist.get_env()
        real_name = lib.get_local_node_name(env.cmd_runner())
        self.assertEqual(expected_name, real_name)

    def test_error(self):
        self.config.runner.pcmk.local_node_name(
            stdout="some info", stderr="some error", returncode=1
        )
        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_name(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND,
                    force_code=None,
                    reason="some error\nsome info",
                )
            ],
            expected_in_processor=False,
        )

    def test_error_not_connected(self):
        stderr = (
            "error: Could not connect to controller: Transport endpoint is "
            "not connected\n"
        )
        self.config.runner.pcmk.local_node_name(
            stderr=stderr,
            returncode=_EXITCODE_NOT_CONNECTED,
        )
        env = self.env_assist.get_env()
        with self.assertRaises(lib.PacemakerNotConnectedException) as cm:
            lib.get_local_node_name(env.cmd_runner())
        assert_report_item_list_equal(
            cm.exception.args,
            [
                (
                    Severity.ERROR,
                    report_codes.PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND,
                    {
                        "reason": stderr.strip(),
                    },
                ),
            ],
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class GetLocalNodeStatusTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.nodes_xml = """
            <nodes>
                <node id="1" name="name_1" />
                <node id="2" name="name_2" />
                <node id="3" name="name_3" />
                <node id="4" name="name_4" />
            </nodes>
        """

    def test_offline_status(self):
        self.config.runner.pcmk.load_state(
            stdout=fixture_crm_mon.error_xml_not_connected(),
            returncode=_EXITCODE_NOT_CONNECTED,
        )

        env = self.env_assist.get_env()
        real_status = lib.get_local_node_status(env.cmd_runner())
        self.assertEqual(dict(offline=True), real_status)

    def test_offline_node_name(self):
        self.config.runner.pcmk.load_state(nodes=self.nodes_xml)
        self.config.runner.pcmk.local_node_name(
            stderr=(
                "error: Could not connect to controller: Transport endpoint is "
                "not connected\n"
            ),
            returncode=_EXITCODE_NOT_CONNECTED,
        )

        env = self.env_assist.get_env()
        real_status = lib.get_local_node_status(env.cmd_runner())
        self.assertEqual(dict(offline=True), real_status)

    def test_invalid_status(self):
        self.config.runner.pcmk.load_state(stdout="invalid xml")

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_FORMAT, force_code=None
                )
            ],
            expected_in_processor=False,
        )

    def test_success(self):
        self.config.runner.pcmk.load_state(
            nodes=self.nodes_xml
        ).runner.pcmk.local_node_name(node_name="name_2")

        env = self.env_assist.get_env()
        real_status = lib.get_local_node_status(env.cmd_runner())
        self.assertEqual("2", real_status.get("id"))

    def test_node_not_in_status(self):
        self.config.runner.pcmk.load_state(
            nodes=self.nodes_xml
        ).runner.pcmk.local_node_name(node_name="name_X")

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    force_code=None,
                    node="name_X",
                    searched_types=[],
                )
            ],
            expected_in_processor=False,
        )

    def test_error_getting_node_name(self):
        self.config.runner.pcmk.load_state(
            nodes=self.nodes_xml
        ).runner.pcmk.local_node_name(
            stdout="some info", stderr="some error", returncode=1
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_local_node_status(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND,
                    force_code=None,
                    reason="some error\nsome info",
                )
            ],
            expected_in_processor=False,
        )


class RemoveNode(TestCase):
    # pylint: disable=no-self-use
    def test_success(self):
        mock_runner = get_runner("", "", 0)
        lib.remove_node(mock_runner, "NODE_NAME")
        mock_runner.run.assert_called_once_with(
            [path("crm_node"), "--force", "--remove", "NODE_NAME"]
        )

    def test_error(self):
        expected_stderr = "expected stderr"
        mock_runner = get_runner("", expected_stderr, 1)
        assert_raise_library_error(
            lambda: lib.remove_node(mock_runner, "NODE_NAME"),
            (
                Severity.ERROR,
                report_codes.NODE_REMOVE_IN_PACEMAKER_FAILED,
                {
                    "node": "",
                    "node_list_to_remove": ["NODE_NAME"],
                    "reason": expected_stderr,
                },
            ),
        )


class ResourceCleanupTest(TestCase):
    def setUp(self):
        self.stdout = "expected output"
        self.stderr = "expected stderr"
        self.resource = "my_resource"
        self.node = "my_node"
        self.env_assist, self.config = get_env_tools(test_case=self)

    def assert_output(self, real_output):
        self.assertEqual(self.stdout + "\n" + self.stderr, real_output)

    def test_basic(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout, stderr=self.stderr
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(env.cmd_runner())
        self.assert_output(real_output)

    def test_resource(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout, stderr=self.stderr, resource=self.resource
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(), resource=self.resource
        )
        self.assert_output(real_output)

    def test_resource_strict(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource,
            strict=True,
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(), resource=self.resource, strict=True
        )
        self.assert_output(real_output)

    def test_node(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout, stderr=self.stderr, node=self.node
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(env.cmd_runner(), node=self.node)
        self.assert_output(real_output)

    def test_all_options(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource,
            node=self.node,
            strict=True,
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_cleanup(
            env.cmd_runner(),
            resource=self.resource,
            node=self.node,
            strict=True,
        )
        self.assert_output(real_output)

    def test_error_cleanup(self):
        self.config.runner.pcmk.resource_cleanup(
            stdout=self.stdout, stderr=self.stderr, returncode=1
        )
        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.resource_cleanup(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.RESOURCE_CLEANUP_ERROR,
                    reason=(self.stderr + "\n" + self.stdout),
                    resource=None,
                    node=None,
                )
            ],
            expected_in_processor=False,
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class ResourceRefreshTest(TestCase):
    def setUp(self):
        self.stdout = "expected output"
        self.stderr = "expected stderr"
        self.resource = "my_resource"
        self.node = "my_node"
        self.env_assist, self.config = get_env_tools(test_case=self)

    def assert_output(self, real_output):
        self.assertEqual(self.stdout + "\n" + self.stderr, real_output)

    @staticmethod
    def fixture_status_xml(nodes, resources):
        doc = etree.parse(rc("crm_mon.minimal.xml"))
        doc.find("/summary/nodes_configured").set("number", str(nodes))
        doc.find("/summary/resources_configured").set("number", str(resources))
        return etree_to_str(doc)

    def test_basic(self):
        self.config.runner.pcmk.load_state(stdout=self.fixture_status_xml(1, 1))
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout, stderr=self.stderr
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(env.cmd_runner())
        self.assert_output(real_output)

    def test_threshold_exceeded(self):
        self.config.runner.pcmk.load_state(
            stdout=self.fixture_status_xml(1000, 1000)
        )
        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.resource_refresh(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.RESOURCE_REFRESH_TOO_TIME_CONSUMING,
                    force_code=report_codes.FORCE,
                    threshold=100,
                )
            ],
            expected_in_processor=False,
        )

    def test_threshold_exceeded_forced(self):
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout, stderr=self.stderr
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(env.cmd_runner(), force=True)
        self.assert_output(real_output)

    def test_resource(self):
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout, stderr=self.stderr, resource=self.resource
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(
            env.cmd_runner(), resource=self.resource
        )
        self.assert_output(real_output)

    def test_resource_strict(self):
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource,
            strict=True,
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(
            env.cmd_runner(), resource=self.resource, strict=True
        )
        self.assert_output(real_output)

    def test_node(self):
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout, stderr=self.stderr, node=self.node
        )
        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(env.cmd_runner(), node=self.node)
        self.assert_output(real_output)

    def test_all_options(self):
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout,
            stderr=self.stderr,
            resource=self.resource,
            node=self.node,
            strict=True,
        )

        env = self.env_assist.get_env()
        real_output = lib.resource_refresh(
            env.cmd_runner(),
            resource=self.resource,
            node=self.node,
            strict=True,
        )
        self.assert_output(real_output)

    def test_error_state(self):
        self.config.runner.pcmk.load_state(
            stdout=self.stdout, stderr=self.stderr, returncode=1
        )
        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.resource_refresh(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason=(self.stderr + "\n" + self.stdout),
                )
            ],
            expected_in_processor=False,
        )

    def test_error_refresh(self):
        self.config.runner.pcmk.load_state(stdout=self.fixture_status_xml(1, 1))
        self.config.runner.pcmk.resource_refresh(
            stdout=self.stdout, stderr=self.stderr, returncode=1
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: lib.resource_refresh(env.cmd_runner()),
            [
                fixture.error(
                    report_codes.RESOURCE_REFRESH_ERROR,
                    reason=(self.stderr + "\n" + self.stdout),
                    resource=None,
                    node=None,
                )
            ],
            expected_in_processor=False,
        )


class ResourcesWaitingTest(TestCase):
    # pylint: disable=no-self-use
    def test_wait_success(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 0
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        self.assertEqual(None, lib.wait_for_idle(mock_runner, 0))

        mock_runner.run.assert_called_once_with(
            [path("crm_resource"), "--wait"]
        )

    def test_wait_timeout_success(self):
        expected_stdout = "expected output"
        expected_stderr = "expected stderr"
        expected_retval = 0
        timeout = 10
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        self.assertEqual(None, lib.wait_for_idle(mock_runner, timeout))

        mock_runner.run.assert_called_once_with(
            [path("crm_resource"), "--wait", "--timeout={0}".format(timeout)]
        )

    def test_wait_error(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 1
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.wait_for_idle(mock_runner, 0),
            (
                Severity.ERROR,
                report_codes.WAIT_FOR_IDLE_ERROR,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [path("crm_resource"), "--wait"]
        )

    def test_wait_error_timeout(self):
        expected_stdout = "some info"
        expected_stderr = "some error"
        expected_retval = 124
        mock_runner = get_runner(
            expected_stdout, expected_stderr, expected_retval
        )

        assert_raise_library_error(
            lambda: lib.wait_for_idle(mock_runner, 0),
            (
                Severity.ERROR,
                report_codes.WAIT_FOR_IDLE_TIMED_OUT,
                {
                    "reason": expected_stderr + "\n" + expected_stdout,
                },
            ),
        )

        mock_runner.run.assert_called_once_with(
            [path("crm_resource"), "--wait"]
        )


class IsInPcmkToolHelp(TestCase):
    # pylint: disable=protected-access
    def test_all_in_stderr(self):
        mock_runner = get_runner("", "ABCDE", 0)
        self.assertTrue(
            lib._is_in_pcmk_tool_help(mock_runner, "", ["A", "C", "E"])
        )

    def test_all_in_stdout(self):
        mock_runner = get_runner("ABCDE", "", 0)
        self.assertTrue(
            lib._is_in_pcmk_tool_help(mock_runner, "", ["A", "C", "E"])
        )

    def test_some_in_stderr_all_in_stdout(self):
        mock_runner = get_runner("ABCDE", "ABC", 0)
        self.assertTrue(
            lib._is_in_pcmk_tool_help(mock_runner, "", ["A", "C", "E"])
        )

    def test_some_in_stderr_some_in_stdout(self):
        mock_runner = get_runner("CDE", "ABC", 0)
        self.assertFalse(
            lib._is_in_pcmk_tool_help(mock_runner, "", ["A", "C", "E"])
        )


class GetRulesInEffectStatus(TestCase):
    def test_success(self):
        test_data = [
            (1, CibRuleInEffectStatus.UNKNOWN),
            (110, CibRuleInEffectStatus.EXPIRED),
            (0, CibRuleInEffectStatus.IN_EFFECT),
            (111, CibRuleInEffectStatus.NOT_YET_IN_EFFECT),
        ]
        for return_code, response in test_data:
            with self.subTest(return_code=return_code, response=response):
                runner = mock.MagicMock(spec_set=CommandRunner)
                runner.run.return_value = ("", "", return_code)
                self.assertEqual(
                    lib.get_rule_in_effect_status(runner, "mock cib", "ruleid"),
                    response,
                )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class GetResourceDigests(TestCase):
    DIGESTS = {
        "all": "0" * 31 + "1",
        "nonprivate": "0" * 31 + "2",
        "nonreloadable": "0" * 31 + "3",
    }
    CALL_ARGS = [
        path("crm_resource"),
        "--digests",
        "--resource",
        "resource-id",
        "--node",
        "node1",
        "--output-as",
        "xml",
        "opt_name1=opt_value1",
        "opt_name2=opt_value2",
        "CRM_meta_interval=30000",
        "CRM_meta_timeout=10000",
    ]
    CRM_ATTRS = {"interval": "30000", "timeout": "10000"}
    RESOURCE_OPTIONS = {"opt_name1": "opt_value1", "opt_name2": "opt_value2"}
    FIXTURE_PACEMAKER_ERROR_XML = """
        <pacemaker-result api-version="2.9" request="crm_resource">
          <status code="102" message="Not connected">
            <errors>
              <error>crm_resource: Could not connect to the CIB: Transport endpoint is not connected
        Error performing operation: Not connected</error>
            </errors>
          </status>
        </pacemaker-result>

    """
    FIXTURE_PACEMAKER_ERROR = (
        "Not connected\ncrm_resource: Could not connect to the CIB: Transport "
        "endpoint is not connected\n        Error performing operation: Not "
        "connected"
    )

    def assert_command_failure(
        self, stdout="", returncode=0, report_output=None
    ):
        if report_output is None:
            report_output = stdout
        runner = get_runner(stdout=stdout, returncode=returncode)
        assert_raise_library_error(
            lambda: lib.get_resource_digests(
                runner,
                "resource-id",
                "node1",
                self.RESOURCE_OPTIONS,
                self.CRM_ATTRS,
            ),
            (
                Severity.ERROR,
                report_codes.UNABLE_TO_GET_RESOURCE_OPERATION_DIGESTS,
                {"output": report_output},
            ),
        )
        runner.run.assert_called_once_with(self.CALL_ARGS)

    def assert_command_sucess(
        self,
        digest_types=("all",),
        resource_options=None,
        crm_attrs=None,
        call_args=None,
    ):
        if resource_options is None:
            resource_options = self.RESOURCE_OPTIONS
        if crm_attrs is None:
            crm_attrs = self.CRM_ATTRS
        if call_args is None:
            call_args = self.CALL_ARGS
        runner = get_runner(stdout=self.fixture_digests_xml(digest_types))
        self.assertEqual(
            lib.get_resource_digests(
                runner,
                "resource-id",
                "node1",
                resource_options,
                crm_meta_attributes=crm_attrs,
            ),
            self.fixture_result_dict(digest_types),
        )
        runner.run.assert_called_once_with(call_args)

    def fixture_result_dict(self, digest_types=()):
        result_dict = {k: None for k in self.DIGESTS}
        for digest_type in digest_types:
            result_dict[digest_type] = self.DIGESTS[digest_type]
        return result_dict

    def fixture_digests_xml(self, digest_types=()):
        return """
            <pacemaker-result api-version="2.9" request="crm_resource">
                <digests resource="resource-id" node="node1" task="monitor" interval="0ms">
                    {digests}
                </digests>
                <status code="0" message="OK"/>
            </pacemaker-result>

        """.format(
            digests="\n".join(
                f"""
                <digest type="{digest_type}" hash="{self.DIGESTS[digest_type]}">
                    <parameters/>
                </digest>
                """
                for digest_type in digest_types
            )
        ).strip()

    def test_success(self):
        digest_types = list(self.DIGESTS.keys())
        test_list = [digest_types[0:1], digest_types[0:2], digest_types]
        for tested_digest_types in test_list:
            with self.subTest(tested_digest_types=tested_digest_types):
                self.assert_command_sucess(digest_types=tested_digest_types)

    def test_success_no_crm_attrs(self):
        self.assert_command_sucess(crm_attrs={}, call_args=self.CALL_ARGS[:-2])

    def test_success_empty_resource_options(self):
        self.assert_command_sucess(
            resource_options={}, crm_attrs={}, call_args=self.CALL_ARGS[:-4]
        )

    def test_invalid_xml(self):
        self.assert_command_failure(stdout="invalid_xml")

    def test_not_valid_with_schema(self):
        self.assert_command_failure(stdout="<xml/>")

    def test_pacemaker_error(self):
        self.assert_command_failure(
            stdout=self.FIXTURE_PACEMAKER_ERROR_XML,
            returncode=101,
            report_output=self.FIXTURE_PACEMAKER_ERROR,
        )

    def test_no_digest_found(self):
        self.assert_command_failure(stdout=self.fixture_digests_xml())



from urllib.parse import parse_qs

from pcs.lib.corosync.live import (
    get_local_corosync_conf as original_get_local_corosync_conf,
)


def print_caption(caption, indent=2, underline="-"):
    print(
        "\n{0}{1}\n{0}{2}".format(
            " " * indent, caption, underline * len(caption)
        )
    )


def print_initialize(spy):
    print_caption(
        "Initialize {0}".format(spy.__class__.__name__),
        indent=0,
        underline="=",
    )


def print_call(spy, name):
    print_caption("{0}: {1}".format(spy.__class__.__name__, name))


def print_line(content):
    print("    {0}".format(content))


def print_long_text(name, potentially_long_text):
    if "\n" not in potentially_long_text:
        print_line("{0}: '{1}'".format(name, potentially_long_text))
    else:
        print_line("{0}: '''\\".format(name))
        for line in potentially_long_text.split("\n"):
            print_line("  {0}".format(line))
        print_line("'''")


class NodeCommunicator:
    def __init__(self, original_node_communicator):
        print_initialize(self)
        self.__communicator = original_node_communicator

    def add_requests(self, request_list):
        print_call(self, "add_requests")
        for request in request_list:
            print_line(request)
            print_line(parse_qs(request.data))
        return self.__communicator.add_requests(request_list)

    def start_loop(self):
        for response in self.__communicator.start_loop():
            print_call(self, "yield response start")
            print_line(response)
            yield response


class Runner:
    def __init__(self, original_runner):
        print_initialize(self)
        self.__runner = original_runner

    def run(
        self, args, stdin_string=None, env_extend=None, binary_output=False
    ):
        print_call(self, "run")
        print_line("args: {0}".format(args))
        if stdin_string:
            print_long_text("stdin_string", stdin_string)
        if env_extend:
            print_line("env_extend: {0}".format(env_extend))
        if binary_output:
            print_line("binary_output: {0}".format(binary_output))
        stdout, stderr, returncode = self.__runner.run(
            args,
            stdin_string,
            env_extend,
            binary_output,
        )
        print_long_text("stdout", stdout)
        print_long_text("stderr", stderr)
        print_line("returncode:{0}".format(returncode))
        return stdout, stderr, returncode


def get_local_corosync_conf():
    print_caption("get_local_corosync_conf", indent=0)
    corosync_conf = original_get_local_corosync_conf()
    for line in corosync_conf.split("\n"):
        print_line(line)
    return corosync_conf



import json
import base64

from pcs_test.tools.command_env.mock_node_communicator import (
    place_multinode_call,
)


class BoothShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def send_config(
        self,
        booth_name,
        config,
        authfile=None,
        authfile_data=None,
        node_labels=None,
        communication_list=None,
        name="http.booth.send_config",
    ):
        data = {
            "config": {
                "name": "{}.conf".format(booth_name),
                "data": config,
            }
        }
        if authfile and authfile_data:
            data["authfile"] = {
                "name": authfile,
                "data": base64.b64encode(authfile_data).decode("utf-8"),
            }
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/booth_set_config",
            param_list=[("data_json", json.dumps(data))],
        )

    def get_config(
        self,
        booth_name,
        config_data=None,
        authfile=None,
        authfile_data=None,
        node_labels=None,
        communication_list=None,
        name="http.booth.get_config",
    ):
        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/booth_get_config",
            param_list=[("name", booth_name)],
            output=json.dumps(
                {
                    "config": {
                        "data": config_data,
                    },
                    "authfile": {
                        "name": authfile,
                        "data": base64.b64encode(authfile_data).decode("utf-8")
                        if authfile_data
                        else None,
                    },
                }
            ),
        )

    def save_files(
        self,
        files_data,
        saved=(),
        existing=(),
        failed=(),
        rewrite_existing=False,
        node_labels=None,
        communication_list=None,
        name="http.booth.save_files",
    ):
        # pylint: disable=too-many-arguments
        param_list = [("data_json", json.dumps(files_data))]
        if rewrite_existing:
            param_list.append(("rewrite_existing", "1"))

        place_multinode_call(
            self.__calls,
            name,
            node_labels,
            communication_list,
            action="remote/booth_save_files",
            param_list=param_list,
            output=json.dumps(
                {
                    "saved": saved,
                    "existing": existing,
                    "failed": failed,
                }
            ),
        )



import os.path

from pcs.daemon import ruby_pcsd, session
from pcs.daemon.app import session as app_session
from pcs.daemon.app.common import BaseHandler
from pcs.daemon.app.sinatra_common import Sinatra
from pcs.daemon.app.ui_common import AjaxMixin, StaticFile


class SinatraGui(app_session.Mixin, Sinatra):
    """
    SinatraGui is base class for handlers which calls the Sinatra GUI functions.
    It adds work with session.
    It adds default GET and POST handlers with hook before Sinatra is called.
    """

    can_use_sinatra = True

    def initialize(self, session_storage, ruby_pcsd_wrapper):
        # pylint: disable=arguments-differ
        app_session.Mixin.initialize(self, session_storage)
        Sinatra.initialize(self, ruby_pcsd_wrapper)
        # add security related headers to all responses
        self.enhance_headers()

    def before_sinatra_use(self):
        pass

    async def handle_sinatra_request(self):
        await self.init_session()
        self.before_sinatra_use()
        if self.can_use_sinatra:
            result = await self.ruby_pcsd_wrapper.request_gui(
                self.request,
                self.session.username,
                self.session.groups,
                self.session.is_authenticated,
            )
            self.send_sinatra_result(result)

    async def get(self, *args, **kwargs):
        del args, kwargs
        await self.handle_sinatra_request()

    async def post(self, *args, **kwargs):
        del args, kwargs
        await self.handle_sinatra_request()


class SinatraGuiProtected(SinatraGui):
    """
    SinatraGuiProtected handles urls that calls the non-ajax Sinatra GUI
    functions. These urls provides real pages.
    """

    def before_sinatra_use(self):
        # sinatra must not have a session at this moment. So the response from
        # sinatra does not contain propriate cookie. Now it is new daemons' job
        # to send this cookies.
        self.sid_to_cookies()

        if not self.session.is_authenticated:
            self.redirect("/login", status=302)  # redirect temporary (302)
            self.can_use_sinatra = False


class SinatraAjaxProtected(SinatraGui, AjaxMixin):
    # pylint: disable=too-many-ancestors
    """
    SinatraAjaxProtected handles urls that calls the ajax Sinatra GUI functions.
    It allows to use this urls only for ajax calls.
    """

    @property
    def is_authorized(self):
        # User is authorized only to perform ajax calls to prevent CSRF attack.
        return self.is_ajax and self.session.is_authenticated

    def before_sinatra_use(self):
        # TODO this is for sinatra compatibility, review it.
        if self.was_sid_in_request_cookies():
            self.put_request_cookies_sid_to_response_cookies_sid()
        if not self.is_authorized:
            raise self.unauthorized()


class Login(SinatraGui, AjaxMixin):
    # pylint: disable=too-many-ancestors
    """
    Login handles urls that cares about loging process.
    It displays a login form (taken from Sinatra), process data sent by the
    form, call an authentication mechanism and uses session for expressing
    a result of the authentication mechanism.
    """

    def before_sinatra_use(self):
        # TODO this is for sinatra compatibility, review it.
        if self.was_sid_in_request_cookies():
            self.put_request_cookies_sid_to_response_cookies_sid()

    async def post(self, *args, **kwargs):
        # This is the way of old (ruby) pcsd. Post login generates a session
        # cookie. No matter if authentication succeeded or failed.
        await self.session_auth_user(
            self.get_body_argument("username"),
            self.get_body_argument("password"),
            sign_rejection=not self.is_ajax,
        )

        if self.session.is_authenticated:
            self.__success_response()
        else:
            self.__failed_response()

    def __success_response(self):
        if self.is_ajax:
            self.write(self.session.ajax_id)
        else:
            self.redirect("/manage", status=303)  # post -> get resource (303)

    def __failed_response(self):
        if self.is_ajax:
            raise self.unauthorized()

        self.redirect("/login", status=303)  # post -> get resource (303)


class Logout(app_session.Mixin, AjaxMixin, BaseHandler):
    """
    Logout handles url for logout. It is used for both ajax and non-ajax
    requests.
    """

    async def get(self, *args, **kwargs):
        del args, kwargs
        await self.init_session()
        self.session_logout()
        self.sid_to_cookies()
        self.enhance_headers()
        if self.is_ajax:
            self.write("OK")
        else:
            self.redirect("/login", status=302)  # redirect temporary (302)


def get_routes(
    session_storage: session.Storage,
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
    public_dir,
):
    ruby_wrapper = dict(ruby_pcsd_wrapper=ruby_pcsd_wrapper)
    sessions = dict(session_storage=session_storage)
    static_path = lambda dir: dict(path=os.path.join(public_dir, dir))
    return [
        (r"/css/(.*)", StaticFile, static_path("css")),
        (r"/js/(.*)", StaticFile, static_path("js")),
        (r"/images/(.*)", StaticFile, static_path("images")),
        (r"/login", Login, {**sessions, **ruby_wrapper}),
        (r"/logout", Logout, sessions),
        # The protection by session was moved from ruby code to python code
        # (tornado).
        (
            r"/($|manage/?$|permissions/?$|managec/.+/main)",
            SinatraGuiProtected,
            {**sessions, **ruby_wrapper},
        ),
        (r"/.*", SinatraAjaxProtected, {**sessions, **ruby_wrapper}),
    ]



from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.xml import etree_to_str

CALL_TYPE_PUSH_CIB = "CALL_TYPE_PUSH_CIB"


class Call:
    type = CALL_TYPE_PUSH_CIB

    def __init__(
        self, cib_xml, custom_cib=False, wait_timeout=-1, exception=None
    ):
        self.cib_xml = cib_xml
        self.custom_cib = custom_cib
        self.wait_timeout = wait_timeout
        self.exception = exception

    def __repr__(self):
        return str("<CibPush wait_timeout='{0}'>").format(self.wait_timeout)


def get_push_cib(call_queue):
    def push_cib(lib_env, custom_cib=None, wait_timeout=-1):
        i, expected_call = call_queue.take(CALL_TYPE_PUSH_CIB)

        if custom_cib is None and expected_call.custom_cib:
            raise AssertionError(
                (
                    "Trying to call env.push_cib (call no. {0}) without "
                    "a custom cib but a custom cib was expected"
                ).format(i)
            )
        if custom_cib is not None and not expected_call.custom_cib:
            raise AssertionError(
                (
                    "Trying to call env.push_cib (call no. {0}) with a custom "
                    "cib but no custom cib was expected"
                ).format(i)
            )

        assert_xml_equal(
            expected_call.cib_xml,
            etree_to_str(lib_env.cib),
            (
                "Trying to call env.push_cib (call no. {0}) but cib in env does"
                " not match\n\n"
            ).format(i),
        )

        if wait_timeout != expected_call.wait_timeout:
            raise AssertionError(
                (
                    "Trying to call env.push_cib (call no. {index}) with "
                    "'wait_timeout' == {real_value} ({real_type}) but it was "
                    "expected 'wait_timeout' == {expected_value} "
                    "({expected_type})"
                ).format(
                    index=i,
                    real_value=wait_timeout,
                    real_type=type(wait_timeout),
                    expected_value=expected_call.wait_timeout,
                    expected_type=type(expected_call.wait_timeout),
                )
            )

        if expected_call.exception:
            raise expected_call.exception

    return push_cib


def is_push_cib_call_in(call_queue):
    return call_queue.has_type(CALL_TYPE_PUSH_CIB)



"""
Common functions used from specific constraint commands.
Functions of this module are not intended to be used for direct call from
client.
"""
from functools import partial

from pcs.lib.cib.constraint import constraint, resource_set
from pcs.lib.cib.tools import get_constraints


def create_with_set(
    tag_name,
    prepare_options,
    env,
    resource_set_list,
    constraint_options,
    resource_in_clone_alowed=False,
    duplication_alowed=False,
    duplicate_check=None,
):
    """
    string tag_name is constraint tag name
    callable prepare_options takes
        cib(Element), options(dict), resource_set_list and return corrected
        options or if options not usable raises error
    env is library environment
    list resource_set_list is description of resource set, for example:
        {"ids": ["A", "B"], "options": {"sequential": "true"}},
    dict constraint_options is base for building attributes of constraint tag
    bool resource_in_clone_alowed flag for allowing to reference id which is
        in tag clone or master
    bool duplication_alowed flag for allowing create duplicate element
    callable duplicate_check takes two elements and decide if they are
        duplicates
    """
    cib = env.get_cib()

    find_valid_resource_id = partial(
        constraint.find_valid_resource_id,
        env.report_processor,
        cib,
        resource_in_clone_alowed,
    )

    constraint_section = get_constraints(cib)
    constraint_element = constraint.create_with_set(
        constraint_section,
        tag_name,
        options=prepare_options(cib, constraint_options, resource_set_list),
        resource_set_list=[
            resource_set.prepare_set(find_valid_resource_id, resource_set_item)
            for resource_set_item in resource_set_list
        ],
    )

    if not duplicate_check:
        duplicate_check = constraint.have_duplicate_resource_sets

    constraint.check_is_without_duplication(
        env.report_processor,
        constraint_section,
        constraint_element,
        are_duplicate=duplicate_check,
        export_element=constraint.export_with_set,
        duplication_alowed=duplication_alowed,
    )

    env.push_cib()


def config(tag_name, is_plain, env):
    """
    string tag_name is constraint tag name
    callable is_plain takes constraint element and returns if is plain (i.e.
        without resource set)
    env is library environment
    """
    constraints_info = {"plain": [], "with_resource_sets": []}
    for element in get_constraints(env.get_cib()).findall(".//" + tag_name):
        if is_plain(element):
            constraints_info["plain"].append(constraint.export_plain(element))
        else:
            constraints_info["with_resource_sets"].append(
                constraint.export_with_set(element)
            )
    return constraints_info
